

set -ex



export CONTINUE_THROUGH_ERROR=1
export MATRIX_GPU_ARCH_VERSION="none"
export MATRIX_GPU_ARCH_TYPE="none"
export MATRIX_CHANNEL="defaults"
export MATRIX_STABLE_VERSION="2.1.0"
export MATRIX_PACKAGE_TYPE="conda"
export TARGET_OS="unix"
python ./smoke_test.py --package torchonly
python ./test/run_test.py --core || true
pip check
python -c "import torch; print(torch.__version__)"
python -c "import torch; import numpy"
python -c "import numpy; import torch"
python -c "import torch; assert torch.distributed.is_available()"
exit 0
